import "foo";
import foo1 from "foo";
import { default as foo2 } from "foo";
import * as foo3 from "foo";
import foo4, { baz as xyz1 } from "foo";
import { bar1 } from "foo";
import { bar2, baz1 } from "foo";
import { bar as baz2 } from "foo";
import { bar as baz3, xyz2 } from "foo";