import type Def1 from "foo";
import type {named1} from "foo";
import type Def2, {named2} from "foo";
import type switch1 from "foo";
import type { switch2 } from "foo";
import type { foo1, bar1 } from "baz";
import type from "foo";
import typeof foo3 from "bar";
import typeof switch4 from "foo";
import typeof { switch5 } from "foo";
import typeof { foo as bar6 } from "baz";
import typeof * as ns7 from "foo";
import typeof * as switch8 from "foo";
