var foo = "bar";

var _bar = babelHelpers.classPrivateFieldLooseKey("bar");

var _baz = babelHelpers.classPrivateFieldLooseKey("baz");

var Foo = function Foo(_foo) {
  "use strict";

  babelHelpers.classCallCheck(this, Foo);
  Object.defineProperty(this, _bar, {
    writable: true,
    value: this
  });
  Object.defineProperty(this, _baz, {
    writable: true,
    value: foo
  });
};
