var foo = "this is a long string";

/**
 * this is comment
 */
class Bar {}

{
  const foo = "this is a really long line of text. this is a really long line of text.";

  /**
   * Example
   */
  class A {}
}