function t1({responseText: responseText}) {
}
function t2({responseText}) {
}
function t3([a, b]) {
}
var [i1, j1, k1] = array;
var {
  i2,
  j2,
  k2
} = obj;
let {i3, j3, k3} = obj;
const {i4, j4, k4} = obj;
var { value1 } = obj;
var {
  value2
} = obj;
