// Newline
`before
after`;

// Newline newline
`before

after`;

// Newline LineContinuation
`before
\
after`;

// LineContinuation
`before\
after`;

// LineContinuation newline
`before\

after`;

// LineContinuation LineContinuation
`before\
\
after`;
