let foo1;
var foo2;

let foo3 = "foo";
var foo4 = "bar";
const foo5 = "foo";

let foo6, bar1 = "bar";
var foo7, bar2 = "bar";

let foo8 = "foo", bar3 = "bar";
var foo9 = "foo", bar4 = "bar";
const foo10 = "foo", bar5 = "bar";
