const nonTrailing = [
  "One", // One
   // Two
  "Two" // Three
  // Four
]

const trailingAfterComma = [
  "One", // One
   // Two
  "Two", // Three
  // Four
]

const trailingAfterArray = [
  "One", // One
   // Two
] // Three
