abstract class C {
    readonly r;
    readonly r2?: number;
    abstract a;
    static s;

    public pu;
    protected po;
    private pi;

    readonly abstract ra;
    abstract readonly ar;
    static readonly sr;

    public readonly pur;
    public abstract pua;
    public static pus;
    public readonly abstract pura;
    public abstract readonly puar;
    public static readonly pusr;
}
