var A = function A(_force) {
  "use strict";

  babelHelpers.classCallCheck(this, A);
  babelHelpers.defineProperty(this, "force", force);
  babelHelpers.defineProperty(this, "foo", babelHelpers.get(babelHelpers.getPrototypeOf(A.prototype), "method", this).call(this));
};
