var createClass = k => {
  var _temp;

  var _k;

  return _temp = (_k = k(), /*#__PURE__*/function () {
    "use strict";

    function _class2() {
      babelHelpers.classCallCheck(this, _class2);
      babelHelpers.defineProperty(this, _k, 2);
    }

    return _class2;
  }()), _temp;
};
