// bounds
type T1 = any;
type T2 = mixed;
type T3 = empty;
// builtins
type T4 = void;
type T5 = number;
type T6 = string;
type T7 = bool;
type T8 = boolean;
// literal type annotations
type T9 = null;
type T10 = "";
type T11 = 0;
type T12 = true;
type T13 = false;
