function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

let Foo = /*#__PURE__*/function () {
  "use strict";

  function Foo() {
    _classCallCheck(this, Foo);
  }

  babelHelpers.createClass(Foo, [{
    key: "method",
    value: function method() {}
  }]);
  return Foo;
}();
