// Copy of modifiers-methods with 'async'
abstract class C {
  abstract async a();

  static async s() {}

  public abstract async pua();

  public static async pus() {}

  public async pu() {}

  protected async po() {}

  private async pi() {}

}