class Demo extends React.Component {
  constructor(props) {
    super(props);
    this.state = {};
  }
  render() {
    return <div>
      <Dialog>
        {/*<div>I am a jsx comment!!</div>*/}
      </Dialog>
    </div>;
  }
}
