class Foo {
  async foo() {}
  foo() {}
  ["foo"]() {}
  get foo() {}
  set foo(bar) {}

  async #foo() {}
  #foo2() {}
  get #foo3() {}
  set #foo3(bar) {}
  * #foo4() {}
  async * #foo5() {}
  get #bar6() {}
  set #baz6(taz) {}

  static async foo() {}
  static foo() {}
  static ["foo"]() {}
  static get foo() {}
  static set foo(bar) {}
  static static() {}
  static * foo() {}
  static async * foo() {}

  static #foo7() {}
  static async #foo8() {}
  static ["foo"]() {}
  static get #foo9() {}
  static set #foo9(taz) {}
  static * #foo10() {}
  static async * #foo11() {}

  get
  () {}

  set
  () {}

  static
  () {}

  async
  () {}

  static
  get
  () {}

  static
  set
  () {}

  static
  static
  () {}

  static
  async
  () {}

  static
  a
  () {}

  get
  async
  () {}

  static
  get
  static
  () {}
}