var _class, _num, _temp;

var Foo = (_temp = (_num = babelHelpers.classPrivateFieldLooseKey("num"), _class = class Foo {}), Object.defineProperty(_class, _num, {
  writable: true,
  value: 0
}), _temp);
