// Copy of modifiers-methods with 'get'
abstract class C {
  abstract get a();

  static get s() {
    return 0;
  }

  public abstract get pua();

  public static get pus() {
    return 0;
  }

  public get pu() {
    return 0;
  }

  protected get po() {
    return 0;
  }

  private get pi() {
    return 0;
  }

}