function _$_renameDeepGlobal_main() { return _$_renameDeepGlobal_dependency() || Promise; }

function _$_renameDeepGlobal_dependency() { return Promise; }

_$_renameDeepGlobal_main;
let _Promise = "I will be a good guy!";
