class A {
  declare static foo;
  static declare foo0: string;

  declare public foo1;
  public declare foo2;

  declare public static foo4;
  public declare static foo3;
  public static declare foo5;
}
