export declare const x: number;
export declare function f(): void;
export declare class C {}
export declare interface I {}
export declare type T = number;
export declare module M {}
export declare namespace N {}
export declare enum foo {}
