var _property = babelHelpers.classPrivateFieldLooseKey("property");

export class MyClass {}
Object.defineProperty(MyClass, _property, {
  writable: true,
  value: value
});

var _property2 = babelHelpers.classPrivateFieldLooseKey("property");

export default class MyClass2 {}
Object.defineProperty(MyClass2, _property2, {
  writable: true,
  value: value
});
