enum E {
  A = true,
  B = false,
}
enum E of boolean {
  A = true,
  B = false,
}
enum E {
  A = 1,
  B = 2,
}
enum E of number {
  A = 1,
  B = 2,
}
enum E {
  A,
  B,
}
enum E of string {
  A,
  B,
}
enum E {
  A = "a",
  B = "b",
}
enum E of string {
  A = "a",
  B = "b",
}
enum E of symbol {
  A,
  B,
}
