// @flow
const a: typeof default = "hi";
const b: typeof stuff.default = "hi";

const c: typeof any = "hi";
const d: typeof bool = "hi";
const e: typeof boolean = "hi";
const f: typeof empty = "hi";
const g: typeof false = "hi";
const h: typeof mixed = "hi";
const i: typeof null = "hi";
const j: typeof number = "hi";
const k: typeof string = "hi";
const l: typeof true = "hi";
const m: typeof void = "hi";
