var foo = "bar";

var Foo = function Foo() {
  "use strict";

  babelHelpers.classCallCheck(this, Foo);
  babelHelpers.defineProperty(this, "bar", foo);
  var _foo = "foo";
  var baz = "baz";
};

babelHelpers.defineProperty(Foo, "bar", baz);
