var _class, _temp;

call((_temp = _class = function _class() {
  babelHelpers.classCallCheck(this, _class);
}, babelHelpers.defineProperty(_class, "test", true), _temp));

var _default = function _default() {
  babelHelpers.classCallCheck(this, _default);
};

babelHelpers.defineProperty(_default, "test", true);
export { _default as default };
;
