import {type} from "foo";
import {type t} from "foo";
import {type as} from "foo";
import {type as as foo} from "foo";
import {type t as u} from "foo";
import {type switch} from "foo";

import {typeof t2} from "foo";
import {typeof as2} from "foo";
import {typeof t as u2} from "foo";
import {typeof switch2} from "foo";
