function foo1(bar = "bar") {}

function foo2(bar = "bar", ...items) {}

function foo3(...items) {}

function* foo4() {}

function foo5() {}

async function foo6() {}

var foo7 = function* foo7() {};

var foo8 = function foo8() {};

var foo9 = async function foo9() {};

var foo10 = function* () {};

var foo11 = function () {};

var foo12 = async function () {};